@@ -28,7 +28,7 @@
 
 package ORM::Db::DBI;
 
-$VERSION = 0.8;
+$VERSION = 0.83;
 
 use DBI;
 use base 'ORM::Db';
@@ -55,20 +55,27 @@ sub new
     my $class = shift;
     my $self  = {};
     my %arg   = @_;
+    my $data_source;
 
     $self->{connect_retries} = defined $arg{connect_retries} ? int( $arg{connect_retries} ) : 3;
     $self->{retry_sleep}     = defined $arg{retry_sleep}     ? int( $arg{retry_sleep} )     : 1;
     $self->{delayed_connect} = $arg{delayed_connect};
     $self->{database}        = $arg{database};
-    $self->{db_arg}          =
-    [
-        "DBI:$arg{driver}:$arg{database}"
+
+    if( $arg{data_source} )
+    {
+        $data_source = $arg{data_source};
+    }
+    else
+    {
+        $data_source =
+            "DBI:$arg{driver}:$arg{database}"
             . ($arg{host}    ? ":$arg{host}" : '')
-            . ($arg{options} ? ";$arg{options}" : ''),
-        $arg{user},
-        $arg{password},
-    ];
-    $self->{db} = DBI->connect( @{$self->{db_arg}} ) unless( $arg{delayed_connect} );
+            . ($arg{options} ? ";$arg{options}" : '');
+    }
+
+    $self->{db_arg} = [ $data_source, $arg{user}, $arg{password} ];
+    $self->{db}     = DBI->connect( @{$self->{db_arg}} ) unless( $arg{delayed_connect} );
 
     return bless $self, $class;
 }
@@ -127,7 +134,7 @@ sub select_base
     {
         for my $name ( keys %{$arg{data}} )
         {
-            $tjoin->merge( $arg{data}{$name}->_tjoin );
+            $tjoin->merge( $arg{data}{$name}->_tjoin ) if( defined $arg{data}{$name} );
         }
         for my $group_by ( @{$arg{group_by}} )
         {
@@ -174,10 +181,10 @@ sub select_base
         $select = '';
         for my $alias ( keys %{$arg{data}} )
         {
+            my $data = ref $arg{data}{$alias} ? $arg{data}{$alias} : ORM::Const->new( $arg{data}{$alias} );
+
             $select .= ",\n" if( $select );
-            $select .=
-                  '  ' . $arg{data}{$alias}->_sql_str( tjoin=>$tjoin )
-                . ' AS ' . $self->qi( $alias );
+            $select .= '  ' . $data->_sql_str( tjoin=>$tjoin ) . ' AS ' . $self->qi( $alias );
         }
     }
     else
@@ -255,7 +262,7 @@ sub select_full
 
         for my $inh_class ( keys %class2id )
         {
-            ORM->_load_ORM_class( $inh_class );
+            $arg{class}->_load_ORM_class( $inh_class );
             %residual_tables = ();
             chop $class2id{ $inh_class };
             for
@@ -350,6 +357,7 @@ sub insert_object
     my $obj_class = ref $obj;
     my $id        = $arg{id};
     my $error     = ORM::Error->new;
+    my $ta        = $obj_class->new_transaction( error=>$error );
 
     # Insert new records into tables
     my @table = $obj_class->_db_tables;
@@ -395,28 +403,6 @@ sub insert_object
         }
     }
 
-    # Roll back if previous loop failed
-    if( $error->fatal )
-    {
-        for( $i=$i-2; $i>=0; $i-- )
-        {
-            my $rows_affected = $self->delete_by_id
-            (
-                table => $table[$i],
-                id    => $id,
-                error => $error,
-            );
-            if( $rows_affected != 1 )
-            {
-                $error->add_fatal
-                (
-                    "Failed to delete row with id#$id from '$table[$i]' during rollback, "
-                    . "$rows_affected rows affected"
-                );
-            }
-        }
-    }
-
     $error->upto( $arg{error} );
     return $id;
 }
@@ -548,26 +534,12 @@ sub delete_object
     my @table     = $obj_class->_db_tables;
     my $rows_affected;
 
-    if( $arg{emulate_foreign_keys} )
-    {
-        for my $ref ( $obj_class->_rev_refs )
-        {
-            my $referers = $ref->[0]->count
-            (
-                filter => ( $ref->[0]->M->_prop($ref->[1])==$obj->id ),
-                error  => $error,
-            );
-            if( $referers )
-            {
-                $error->add_fatal
-                (
-                    "Can't delete instance ID#" . $obj->id
-                    . " of '$obj_class', because there're "
-                    . "$referers instances of '$ref->[0]' refer to it."
-                );
-            }
-        }
-    }
+    $self->check_object_referers
+    (
+        object => $obj,
+        error  => $error,
+        check  => $arg{emulate_foreign_keys},
+    );
 
     unless( $error->fatal )
     {
@@ -584,6 +556,17 @@ sub delete_object
                 $error->add_fatal( "Failed to delete row with id#$id from '$table[$i]' during object delete" );
             }
         }
+
+        # must check twise, new referers could be created during deletion
+        unless( $error->fatal )
+        {
+            $self->check_object_referers
+            (
+                object => $obj,
+                error  => $error,
+                check  => $arg{emulate_foreign_keys},
+            );
+        }
     }
 
     $error->upto( $arg{error} );
@@ -614,7 +597,7 @@ sub referencing_classes
     (
         error => $error,
         query =>
-            'SELECT class,prop FROM _ORM_refs WHERE ref_class='
+            'SELECT class,prop FROM '.$self->qt('_ORM_refs').' WHERE ref_class='
             . $self->qc( $arg{class} )
     );
 
@@ -634,30 +617,45 @@ sub begin_transaction
 {
     my $self  = shift;
     my %arg   = @_;
+    my $error = ORM::Error->new;
 
     $self->{ta} = 1;
-    $self->do( query=>"START TRANSACTION", error=>$arg{error} );
+    $self->_db_handler->begin_work();
+    $error->add_fatal( $self->_db_handler->errstr ) if( $self->_db_handler->err );
+    ORM::DbLog->new( sql=>"BEGIN", error=>$error->text );
+
+    $error->upto( $arg{error} );
 }
 
 sub commit_transaction
 {
     my $self  = shift;
     my %arg   = @_;
+    my $error = ORM::Error->new;
 
     delete $self->{ta};
-    $self->do( query=>"COMMIT", error=>$arg{error} );
+    $self->_db_handler->commit();
+    $error->add_fatal( $self->_db_handler->errstr ) if( $self->_db_handler->err );
+    ORM::DbLog->new( sql=>"COMMIT", error=>$error->text );
+
+    $error->upto( $arg{error} );
 }
 
 sub rollback_transaction
 {
     my $self  = shift;
     my %arg   = @_;
+    my $error = ORM::Error->new;
 
     delete $self->{ta};
     unless( $self->{lost_connection} )
     {
-        $self->do( query=>"ROLLBACK", error=>$arg{error} );
+        $self->_db_handler->rollback();
+        $error->add_fatal( $self->_db_handler->errstr ) if( $self->_db_handler->err );
+        ORM::DbLog->new( sql=>"ROLLBACK", error=>$error->text );
     }
+
+    $error->upto( $arg{error} );
 }
 
 ##
@@ -790,7 +788,7 @@ sub select
             $error->add_fatal( "Failed to execute query, 'prepare' returned undef, query='$query'" );
         }
 
-        if( $self->_lost_connection( $st && $st->err ) )
+        if( $st && $st->err && $self->_lost_connection( $st->err ) )
         {
             $self->_db_reconnect;
             $retry = $tries--;
@@ -870,6 +868,37 @@ sub _sql_limit
     return $sql;
 }
 
+sub check_object_referers
+{
+    my $self      = shift;
+    my %arg       = @_;
+    my $obj       = $arg{object};
+    my $obj_class = ref $obj;
+    
+    if( $arg{check} )
+    {
+        for my $ref ( $obj_class->_rev_refs )
+        {
+            my $referers = $ref->[0]->count
+            (
+                filter => ( $ref->[0]->M->_prop($ref->[1])==$obj->id ),
+                error  => $arg{error},
+            );
+            if( $referers )
+            {
+                $arg{error}->add_fatal
+                (
+                    "Can't delete instance ID#" . $obj->id
+                    . " of '$obj_class', because there're "
+                    . "$referers instances of '$ref->[0]' refer to it."
+                );
+            }
+        }
+    }
+
+    return undef;
+}
+
 ##
 ## ABSTRACT METHODS
 ##
@@ -888,4 +917,4 @@ sub _lost_connection
 ## SQL FUNCTIONS
 ##
 
-sub _func_concat        { shift; ORM::Filter::Func->new( 'CONCAT', @_ ); }
+sub _func_concat { shift; ORM::Filter::Cmp->new( '||', @_ ); }
